@@ -2,7 +2,7 @@ use std::sync::Arc;
 
 use crate::{
     ast,
-    error::{IntErr, Interrupt},
+    error::{FendError, Interrupt},
     lexer, parser,
     scope::Scope,
     value::Value,
@@ -14,14 +14,14 @@ pub(crate) fn evaluate_to_value<'a, I: Interrupt>(
     scope: Option<Arc<Scope>>,
     context: &mut crate::Context,
     int: &I,
-) -> Result<Value, IntErr<String, I>> {
+) -> Result<Value, FendError> {
     let lex = lexer::lex(input, int);
     let mut tokens = vec![];
     let mut missing_open_parens: i32 = 0;
     for token in lex {
-        let token = token.map_err(IntErr::into_string)?;
+        let token = token?;
         if let lexer::Token::Symbol(lexer::Symbol::CloseParens) = token {
-            missing_open_parens += 1
+            missing_open_parens += 1;
         }
         tokens.push(token);
     }
@@ -38,7 +38,7 @@ pub(crate) fn evaluate_to_spans<'a, I: Interrupt>(
     scope: Option<Arc<Scope>>,
     context: &mut crate::Context,
     int: &I,
-) -> Result<Vec<Span>, IntErr<String, I>> {
+) -> Result<Vec<Span>, FendError> {
     let debug = input.strip_prefix("!debug ").map_or(false, |remaining| {
         input = remaining;
         true
@@ -48,7 +48,7 @@ pub(crate) fn evaluate_to_spans<'a, I: Interrupt>(
         vec![Span::from_string(format!("{:?}", value))]
     } else {
         let mut spans = vec![];
-        value.format(0, &mut spans, int)?;
+        value.format(0, &mut spans, context, int)?;
         spans
     })
 }
