@@ -1,12 +1,12 @@
 #![forbid(unsafe_code)]
-#![forbid(clippy::all)]
+#![deny(clippy::all)]
 #![deny(clippy::pedantic)]
 #![deny(clippy::use_self)]
 #![forbid(clippy::needless_borrow)]
 //#![forbid(clippy::cognitive_complexity)]
 #![forbid(unreachable_pub)]
 #![forbid(elided_lifetimes_in_paths)]
-#![doc(html_root_url = "https://docs.rs/fend-core/0.1.20")]
+#![doc(html_root_url = "https://docs.rs/fend-core/0.1.23")]
 
 mod ast;
 mod date;
@@ -116,6 +116,12 @@ enum FCMode {
     CoulombFarad,
 }
 
+#[derive(Clone, Debug, PartialEq, Eq)]
+enum OutputMode {
+    SimpleText,
+    TerminalFixedWidth,
+}
+
 /// This struct contains context used for `fend`. It should only be created once
 /// at startup.
 #[derive(Clone, Debug)]
@@ -123,6 +129,8 @@ pub struct Context {
     current_time: Option<CurrentTimeInfo>,
     variables: HashMap<String, value::Value>,
     fc_mode: FCMode,
+    random_u32: Option<fn() -> u32>,
+    output_mode: OutputMode,
 }
 
 impl Default for Context {
@@ -140,6 +148,8 @@ impl Context {
             current_time: None,
             variables: HashMap::new(),
             fc_mode: FCMode::CelsiusFahrenheit,
+            random_u32: None,
+            output_mode: OutputMode::SimpleText,
         }
     }
 
@@ -166,6 +176,22 @@ impl Context {
     pub fn use_coulomb_and_farad(&mut self) {
         self.fc_mode = FCMode::CoulombFarad;
     }
+
+    /// Set a random number generator
+    pub fn set_random_u32_fn(&mut self, random_u32: fn() -> u32) {
+        self.random_u32 = Some(random_u32);
+    }
+
+    /// Clear the random number generator after setting it with via [`Self::set_random_u32_fn`]
+    pub fn disable_rng(&mut self) {
+        self.random_u32 = None;
+    }
+
+    /// Change the output mode fixed-width terminal style. This enables ASCII
+    /// graphs in the output.
+    pub fn set_output_mode_terminal(&mut self) {
+        self.output_mode = OutputMode::TerminalFixedWidth;
+    }
 }
 
 /// This function evaluates a string using the given context. Any evaluation using this
@@ -202,9 +228,7 @@ pub fn evaluate_with_interrupt(
     }
     let result = match eval::evaluate_to_spans(input, None, context, int) {
         Ok(value) => value,
-        // TODO: handle different interrupt values
-        Err(error::IntErr::Interrupt(_)) => return Err("interrupted".to_string()),
-        Err(error::IntErr::Error(e)) => return Err(e),
+        Err(e) => return Err(e.to_string()),
     };
     let mut plain_result = String::new();
     for s in &result {
@@ -217,7 +241,7 @@ pub fn evaluate_with_interrupt(
 }
 
 const fn get_version_as_str() -> &'static str {
-    "0.1.20"
+    "0.1.23"
 }
 
 /// Returns the current version of `fend-core`.
