@@ -302,6 +302,7 @@ const COMMON_SI_DERIVED_UNITS: &[UnitTuple] = &[
     ("coulomb", "", "l@A s", "charge"),
     ("volt", "volts", "l@W/A", "potential difference"),
     ("V", "", "s@volt", ""),
+    ("Ah", "", "s@ampere hour", ""),
     ("ohm", "ohms", "l@V/A", "electrical resistance"),
     ("siemens", "", "l@A/V", "electrical conductance"),
     ("S", "", "s@siemens", ""),
@@ -318,7 +319,8 @@ const COMMON_SI_DERIVED_UNITS: &[UnitTuple] = &[
     ("nt", "", "nit", ""),
     ("lumen", "lumens", "l@cd sr", "luminous flux"),
     ("lm", "", "s@lumen", ""),
-    ("lx", "lux", "l@lm/m^2", "illuminance"),
+    ("lux", "", "l@lm/m^2", "illuminance"),
+    ("lx", "", "lux", "illuminance"),
     ("phot", "phots", "l@1e4 lx", ""),
     ("ph", "", "s@phot", ""),
     ("becquerel", "becquerels", "l@/s", "radioactivity"),
